""" Contains upgrade tasks that are executed when the application is being
upgraded on the server. See :class:`onegov.core.upgrade.upgrade_task`.

"""
from depot.io.utils import FileIntent
from io import BytesIO
from onegov.core.crypto import random_token
from onegov.core.upgrade import upgrade_task
from onegov.core.utils import dictionary_to_binary
from onegov.form import FormDefinitionCollection
from onegov.form import FormFile
from onegov.form import FormSubmission
from onegov.core.orm.types import JSON
from sqlalchemy import Column, Text


@upgrade_task('Enable external form submissions')
def enable_external_form_submissions(context):

    context.operations.alter_column('submissions', 'name', nullable=True)


@upgrade_task('Set payment method for existing forms')
def set_payment_method_for_existing_forms(context):
    forms = FormDefinitionCollection(context.session)

    for form in forms.query():
        form.payment_method = 'manual'


@upgrade_task('Migrate form submission files to onegov.file')
def migrate_form_submission_files_to_onegov_file(context):
    submission_ids = [
        row[0] for row in
        context.session.execute("""
            SELECT submission_id
            FROM submission_files
        """)
    ]

    if not submission_ids:
        return

    submissions = {
        s.id: s for s in
        context.session.query(FormSubmission).filter(
            FormSubmission.id.in_(submission_ids)
        )
    }

    for row in context.session.execute("""
        SELECT submission_id, field_id, filedata
        FROM submission_files
    """):
        submission_id, field_id, filedata = row
        submission = submissions[submission_id]

        replacement = FormFile(
            id=random_token(),
            name=submission.data[field_id]['filename'],
            note=field_id,
            reference=FileIntent(
                BytesIO(dictionary_to_binary({'data': filedata})),
                submission.data[field_id]['filename'],
                submission.data[field_id]['mimetype'],
            )
        )

        assert submission.data[field_id]['data'].startswith('@')

        submission.data[field_id]['data'] = '@' + replacement.id
        submission.data.changed()

        submission.files.append(replacement)

    context.session.flush()
    context.session.execute("DROP TABLE IF EXISTS submission_files")


@upgrade_task('Add payment_method to definitions and submissions')
def add_payment_method_to_definitions_and_submissions(context):

    context.add_column_with_defaults(
        table='forms',
        column=Column('payment_method', Text, nullable=False),
        default=lambda form: form.content.get('payment_method', 'manual')
    )

    context.add_column_with_defaults(
        table='submissions',
        column=Column('payment_method', Text, nullable=False),
        default=lambda submission: (
            submission.form and
            submission.form.content.get('payment_method', 'manual') or
            'manual'
        )
    )

    for form in context.records_per_table('forms'):
        form.content.pop('payment_method', None)


@upgrade_task('Add meta dictionary to submissions')
def add_meta_directory_to_submissions(context):

    context.add_column_with_defaults(
        table='submissions',
        column=Column('meta', JSON, nullable=False),
        default=lambda submission: dict()
    )



def as_float(value):
    return value and float(value) or 0.0


def strip_whitespace(value):
    return value and value.strip(' \r\n') or None


def yubikey_identifier(value):
    return value and value[:12] or ''



import inspect
import weakref

from collections import OrderedDict
from decimal import Decimal
from itertools import groupby
from onegov.form import utils
from onegov.pay import Price
from operator import itemgetter
from wtforms import Form as BaseForm
from wtforms.fields.html5 import EmailField
from wtforms.validators import InputRequired, DataRequired, Optional
from wtforms_components import If, Chain


class Form(BaseForm):
    """ Extends wtforms.Form with useful methods and integrations needed in
    OneGov applications.

    **Fieldsets**

    This form supports fieldsets (which WTForms doesn't recognize). To put
    fields into a fieldset, add a fieldset attribute to the field during
    class definition::

        class MyForm(Form):
            first_name = StringField('First Name', fieldset='Name')
            last_name = StringField('Last Name', fieldset='Name')
            comment = StringField('Comment')

    A form created like this will have two fieldsets, one visible fieldset
    with the legend set to 'Name' and one invisible fieldset containing
    'comment'.

    Fieldsets with the same name are *not* automatically grouped together.
    Instead, fields are taken in the order they are defined and put into the
    same fieldset, if the previous fieldset has the same name.

    That is to say, in this example, we get three fieldsets::

        class MyForm(Form):
            a = StringField('A', fieldset='1')
            b = StringField('B', fieldset='2')
            c = StringField('C', fieldset='1')

    The first fieldset has the label '1' and it contains 'a'. The second
    fieldset has the label '2' and it contains 'b'. The third fieldset has
    the label '3' and it contains 'c'.

    This ensures that all fields are in either a visible or an invisible
    fieldset (see :meth:`Fieldset.is_visible`).

    **Dependencies**

    This form also supports dependencies. So field b may depend on field a, if
    field a has a certain value, field b is shown on the form (with some
    javascript) and its validators are actually executed. If field a does
    not have the required value, field b is hidden with javascript and its
    validators are not executed.

    The validators which are skipped are only the validators passed with the
    field, the validators on the field itself are still invoked (we can't
    skip them). However, only if the optional field is not empty. That is we
    prevent invalid values no matter what, but we allow for empty values if
    the dependent field does not have the required value.

    This sounds a lot more complicated than it is::

        class MyForm(Form):

            option = RadioField('Option', choices=[
                ('yes', 'Yes'),
                ('no', 'No'),
            ])
            only_if_no = StringField(
                label='Only Shown When No',
                validators=[InputRequired()],
                depends_on=('option', 'no')
            )

    **Pricing**

    Pricing is a way to attach prices to certain form fields. A total price
    is calcualted depending on the selections the user makes::

        class MyForm(Form):

            ticket_insurance = RadioField('Option', choices=[
                ('yes', 'Yes'),
                ('no', 'No')
            ], pricing={
                'yes': (10.0, 'CHF')
            })

            discount_code = TextField('Discount Code', pricing={
                'CAMPAIGN2017': (-5.0, 'CHF')
            })

    Note that the pricing has no implicit meaning. This is simply a way to
    attach prices and to get the total through the ``prices()`` and ``total()``
    calls. What you do with these prices is up to you.

    """

    def __init__(self, *args, **kwargs):

        # preprocessors are generators which yield control to give the
        # constructor the chance to call the parent constructor. Their
        # purpose is to handle custom attributes passed to the fields,
        # removing them in the process (so wtforms doesn't trip up).
        preprocessors = [
            self.process_fieldset(),
            self.process_depends_on(),
            self.process_pricing()
        ]

        for processor in preprocessors:
            next(processor)

        super().__init__(*args, **kwargs)

        for processor in preprocessors:
            next(processor, None)

    @classmethod
    def clone(cls):
        """ Creates an independent copy of the form class.

        The fields of the so called class may be manipulated without affecting
        the original class.

        """

        class ClonedForm(cls):
            pass

        for key, unbound_field in cls._unbound_fields:
            setattr(ClonedForm, key, unbound_field.field_class(
                *unbound_field.args,
                **unbound_field.kwargs
            ))

        return ClonedForm

    def process_fieldset(self):
        """ Processes the fieldset parameter on the fields, which puts
        fields into fieldsets.

        In the process the fields are altered so that wtforms recognizes them
        again (that is, attributes only known to us are removed).

        See :class:`Form` for more information.

        """

        self.fieldsets = []

        # consume the fieldset attribute of all unbound fields, as WTForms
        # doesn't know it -> move it to the field which is a *class* attribute
        # (so this only happens once per class)
        for field_id, field in self._unbound_fields:
            if not hasattr(field, 'fieldset'):
                field.fieldset = field.kwargs.pop('fieldset', None)

        fields_by_fieldset = [
            (field.fieldset, field_id)
            for field_id, field in self._unbound_fields
        ]

        # yield control to the constructor so it can call the parent
        yield

        # wtforms' constructor might add more fields not available as
        # unbound fields (like the csrf token)
        if len(self._fields) != len(self._unbound_fields):
            processed = set(f[1] for f in fields_by_fieldset)
            extra = (
                f[1] for f in self._fields.items() if f[0] not in processed
            )
            self.fieldsets.append(Fieldset(None, fields=extra))

        for label, fields in groupby(fields_by_fieldset, key=itemgetter(0)):
            self.fieldsets.append(Fieldset(
                label=label,
                fields=(self._fields[f[1]] for f in fields)
            ))

    def process_depends_on(self):
        """ Processes the depends_on parameter on the fields, which adds the
        ability to have fields depend on values of other fields.

        In the process the fields are altered so that wtforms recognizes them
        again (that is, attributes only known to us are removed).

        See :class:`Form` for more information.

        """

        for field_id, field in self._unbound_fields:

            if 'depends_on' not in field.kwargs:
                continue

            depends_on = field.kwargs.pop('depends_on')

            if not depends_on:
                continue

            field.depends_on = FieldDependency(*depends_on)

            if 'validators' in field.kwargs:
                field.kwargs['validators'] = (
                    If(
                        field.depends_on.fulfilled,
                        Chain(field.kwargs['validators'])
                    ),
                    If(
                        field.depends_on.unfulfilled,
                        Optional()
                    ),
                )

            field.kwargs['render_kw'] = field.kwargs.get('render_kw', {})
            field.kwargs['render_kw'].update(field.depends_on.html_data)

        yield

    def process_pricing(self):
        """ Processes the pricing parameter on the fields, which adds the
        ability to have fields associated with a price.

        See :class:`Form` for more information.

        """

        pricings = {}

        # move the pricing rule to the field class (happens once per class)
        for field_id, field in self._unbound_fields:
            if not hasattr(field, 'pricing'):
                field.pricing = field.kwargs.pop('pricing', None)

        # prepare the pricing rules
        for field_id, field in self._unbound_fields:
            if field.pricing:
                pricings[field_id] = Pricing(field.pricing)

        yield

        # attach the pricing rules to the field instances
        for field_id, pricing in pricings.items():
            self._fields[field_id].pricing = pricing

    def is_visible_through_dependencies(self, field_id):
        """ Returns true if the given field id has visible because all of
        it's parents are visible. A field is invisible if its dependency is
        not met.

        """

        unbound_field = getattr(self.__class__, field_id)
        depends_on = getattr(unbound_field, 'depends_on', None)

        if not depends_on:
            return True

        bound_field = getattr(self, field_id)

        if not depends_on.fulfilled(self, bound_field):
            return False

        return all(
            self.is_visible_through_dependencies(d['field_id'])
            for d in depends_on.dependencies
        )

    def prices(self):
        """ Returns the prices of all selected items depending on the
        formdata. """

        prices = []

        for field_id, field in self._fields.items():
            if not hasattr(field, 'pricing'):
                continue

            if not self.is_visible_through_dependencies(field_id):
                continue

            price = field.pricing.price(field)

            if price is not None:
                prices.append((field_id, price))

        currencies = set(price.currency for _, price in prices)
        assert len(currencies) <= 1, "Mixed currencies are not supported"

        return prices

    def total(self):
        """ Returns the total amount of all prices. """
        prices = self.prices()

        if not prices:
            return None

        return Price(
            sum(price.amount for field_id, price in prices),
            prices[0][1].currency
        )

    def submitted(self, request):
        """ Returns true if the given request is a successful post request. """
        return request.POST and self.validate()

    def ignore_csrf_error(self):
        """ Removes the csrf error from the form if found, after validation.

        Use this only if you know what you are doing (really, never).

        """
        if self.meta.csrf_field_name in self.errors:
            del self.errors[self.meta.csrf_field_name]
            self.csrf_token.errors = []

    @property
    def has_required_email_field(self):
        """ Returns True if the form has a required e-mail field. """
        matches = self.match_fields(
            include_classes=(EmailField, ),
            required=True,
            limit=1
        )

        return matches and True or False

    def match_fields(self, include_classes=None, exclude_classes=None,
                     required=None, limit=None):
        """ Returns field ids matching the given search criteria.

        :include_classes:
            A list of field classes which should be included.

        :excluded_classes:
            A list of field classes which should be excluded.

        :required:
            True if required fields only, False if no required fields.

        :limit:
            If > 0, limits the number of returned elements.

        All parameters may be set to None disable matching it to anything.

        """

        matches = []

        for field_id, field in self._fields.items():
            if include_classes:
                for cls in include_classes:
                    if isinstance(field, cls):
                        break
                else:
                    continue

            if exclude_classes:
                for cls in exclude_classes:
                    if not isinstance(field, cls):
                        break
                else:
                    continue

            if required is True and not self.is_required(field_id):
                continue

            if required is False and self.is_required(field_id):
                continue

            matches.append(field_id)

            if limit and len(matches) == limit:
                break

        return matches

    def is_required(self, field_id):
        """ Returns true if the given field_id is required. """

        for validator in self._fields[field_id].validators:
            if isinstance(validator, (InputRequired, DataRequired)):
                return True
        return False

    def get_useful_data(self, exclude={'csrf_token'}):
        """ Returns the form data in a dictionary, by default excluding data
        that should not be stored in the db backend.

        """

        return {k: v for k, v in self.data.items() if k not in exclude}

    def populate_obj(self, obj, exclude=None, include=None):
        """ A reimplementation of wtforms populate_obj function with the addage
        of optional include/exclude filters.

        If neither exclude nor include is passed, the function works like it
        does in wtforms. Otherwise fields are considered which are included
        but not excluded.

        """

        include = include or set(self._fields.keys())
        exclude = exclude or set()

        for name, field in self._fields.items():
            if name in include and name not in exclude:
                field.populate_obj(obj, name)

    def process(self, *args, **kwargs):
        """ Calls :meth:`process_obj` if ``process()`` was called with
        the ``obj`` keyword argument.

        This saves an extra check in many cases where we want to extend the
        process function, but only *if* an obj has been provided.

        """
        super().process(*args, **kwargs)

        if 'obj' in kwargs:
            self.process_obj(kwargs.get('obj'))

    def process_obj(self, obj):
        """ Called by :meth:`process` if an object was passed.

        Do *not* use this function directly. To process an object, you should
        call ``form.process(obj=obj)`` instead.

        """
        pass

    def delete_field(self, fieldname):
        """ Removes the given field from the form and all the fieldsets. """

        def fieldsets_without_field():
            for fieldset in self.fieldsets:
                if fieldname in fieldset.fields:
                    del fieldset.fields[fieldname]

                if fieldset.fields:
                    yield fieldset

        self.fieldsets = list(fieldsets_without_field())

        del self[fieldname]

    def validate(self):
        """ Adds support for 'ensurances' to the form. An ensurance is a
        method which is called during validation when all the fields have
        been populated. Therefore it is a good place to validate against
        multiple fields.

        All methods which start with ``ensure_`` are ensurances. If and only if
        an ensurance returns False it is considered to have failed. In this
        case the validate method returns False as well. If None or '' or
        any other falsy value is returned, no error is assumed! This avoids
        having to return an extra True at the end of each ensurance.

        When one ensurance fails, the others are still run. Also, there's no
        error display mechanism. Showing an error is left to the ensurance
        itself. It can do so by adding messages to the various error lists
        of the form or by showing an alert through the request.

        """
        result = super().validate()

        for ensurance in self.ensurances:
            if ensurance() is False:
                result = False

        return result

    @property
    def ensurances(self):
        """ Returns the ensurances that need to be checked when validating.

        This property may be overridden if only a subset of all ensurances
        should actually be enforced.

        """

        # inspect.getmembers is no good here as it triggers the properties
        for name in dir(self):
            if name.startswith('ensure_'):

                if isinstance(getattr(type(self), name), property):
                    continue

                member = getattr(self, name)

                if callable(member):
                    yield member


class Fieldset(object):
    """ Defines a fieldset with a list of fields. """

    def __init__(self, label, fields):
        """ Initializes the Fieldset.

        :label: Label of the fieldset (None if it's an invisible fieldset)
        :fields: Iterator of bound fields. Fieldset creates a list of weak
        references to these fields, as they are defined elsewhere and should
        not be kept in memory just because a Fieldset references them.

        """
        self.label = label
        self.fields = OrderedDict((f.id, weakref.proxy(f)) for f in fields)

    def __len__(self):
        return len(self.fields)

    def __getitem__(self, key):
        return self.fields[key]

    @property
    def is_visible(self):
        return self.label is not None

    @property
    def non_empty_fields(self):
        """ Returns only the fields which are not empty. """
        return OrderedDict(
            (id, field) for id, field in self.fields.items() if field.data)


def with_options(widget, **render_options):
    """ Takes a widget class or instance and returns a child-instance of the
    widget class, with the given options set on the render call.

    This makes it easy to use existing WTForms widgets with custom render
    options:

    field = StringField(widget=with_options(TextArea, class_="markdown"))

    Note: With wtforms 2.1 this is no longer necssary. Instead use the
    render_kw parameter of the field class. This function will be deprecated
    in a future release.

    """

    if inspect.isclass(widget):
        class Widget(widget):

            def __call__(self, *args, **kwargs):
                render_options.update(kwargs)
                return super().__call__(*args, **render_options)

        return Widget()
    else:
        class Widget(widget.__class__):

            def __init__(self):
                self.__dict__.update(widget.__dict__)

            def __call__(self, *args, **kwargs):
                render_options.update(kwargs)
                return widget.__call__(*args, **render_options)

        return Widget()


class FieldDependency(object):
    """ Defines a dependency to a field. The given field(s) must have the given
    choice for this dependency to be fulfilled.

    It's possible to depend on NOT the given value by preceeding it with a '!':

        FieldDependency('field_1', '!choice_1')

    To depend on more than one field, add the field_id's and choices to the
    constructor:

        FieldDependency('field_1', 'choice_1')
        FieldDependency('field_1', 'choice_1', 'field_2', 'choice_2')

    """

    def __init__(self, *kwargs):
        assert len(kwargs) and not len(kwargs) % 2

        self.dependencies = []
        for index in range(len(kwargs) // 2):
            choice = kwargs[2 * index + 1]
            invert = choice.startswith('!')
            self.dependencies.append({
                'field_id': kwargs[2 * index],
                'raw_choice': choice,
                'invert': invert,
                'choice': choice[1:] if invert else choice,
            })

    def fulfilled(self, form, field):
        result = True
        for dependency in self.dependencies:
            data = getattr(form, dependency['field_id']).data
            choice = dependency['choice']
            invert = dependency['invert']
            result = result and ((data == choice) ^ invert)
        return result

    def unfulfilled(self, form, field):
        return not self.fulfilled(form, field)

    @property
    def field_id(self):
        return self.dependencies[0]['field_id']

    @property
    def html_data(self):
        value = ';'.join((
            '/'.join((dep['field_id'], dep['raw_choice']))
            for dep in self.dependencies
        ))
        return {'data-depends-on': value}


class Pricing(object):
    """ Defines pricing on a field, returning the correct price for the field
    depending on its rule.

    """

    def __init__(self, rules):
        self.rules = {
            rule: Price(Decimal(value), currency)
            for rule, (value, currency) in rules.items()
        }

    def price(self, field):
        if isinstance(field.data, list):
            total = None

            for value in field.data:
                price = self.rules.get(value, None)

                if price is not None:
                    total = (total or Decimal(0)) + price.amount
                    currency = price.currency

            if total is None:
                return None
            else:
                return Price(total, currency)

        return self.rules.get(field.data, None)


def merge_forms(*forms):
    """ Takes a list of forms and merges them.

    In doing so, a new class is created which inherits from all the forms in
    the default method resolution order. So the first class will override
    fields in the second class and so on.

    So this method is basically the same as:

        class Merged(*forms):
            pass

    With *one crucial difference*, the order of the fields is as follows:

    First, the fields from the first form are rendered, then the fields
    from the second form and so on. This is not the case if you merge the
    forms by simple class inheritance, as each form has it's own internal
    field order, which when merged leads to unexpected results.

    """

    class MergedForm(*forms):
        pass

    fields_in_order = (
        name for cls in forms for name, field
        in utils.get_fields_from_class(cls)
    )

    return enforce_order(MergedForm, fields_in_order)


def enforce_order(form_class, fields_in_order):
    """ Takes a list of fields used in a form_class and enforces the
    order of those fields.

    If not all fields in the form are given, the resulting order is undefined.

    """

    # XXX to make sure the field order of the existing class remains
    # unchanged, we need to instantiate the class once (wtforms seems
    # to do some housekeeping somehwere)
    form_class()

    class EnforcedOrderForm(form_class):
        pass

    processed = set()

    for counter, name in enumerate(fields_in_order, start=1):
        if name in processed:
            continue

        getattr(EnforcedOrderForm, name).creation_counter = counter
        processed.add(name)

    return EnforcedOrderForm


def move_fields(form_class, fields, after):
    """ Reorders the given fields (given by name) by inserting them directly
    after the given field.

    If ``after`` is None, the fields are moved to the end.

    """

    fields_in_order = []

    for name, _ in utils.get_fields_from_class(form_class):
        if name in fields:
            continue

        fields_in_order.append(name)

        if name == after:
            fields_in_order.extend(fields)

    if after is None:
        fields_in_order.extend(fields)

    return enforce_order(form_class, fields_in_order)



form_extensions = {}


class FormExtension(object):
    """ Enables the extension of form definitions/submissions.

    When either of those models create a form class they will take the
    'extensions' key in the meta dictionary to extend those formcode
    based forms.

    This allows for specialised behaviour of formcode forms with the drawback
    that those definitions/submissions are more tightly bound to the code. That
    is to say code in module A could not use submissions defined by module B
    unless module B is also present in the path.

    To create and register a form extension subclass as follows::

        class MyExtension(FormExtension, name='my-extension'):
            def create(self):
                return self.form_class

    Note that you *should not* change the form_class provided to you. Instead
    you should subclass it. If you need to change the form class, you need
    to clone it::

        class MyExtension(FormExtension, name='my-extension'):
            def create(self):
                return self.form_class.clone()

        class MyExtension(FormExtension, name='my-extension'):
            def create(self):
                class ExtendedForm(self.form_class):
                    pass

                return ExtendedForm

    Also, names must be unique and can only be registered once.

    """

    def __init__(self, form_class):
        self.form_class = form_class

    def __init_subclass__(cls, name, **kwargs):
        super().__init_subclass__(**kwargs)

        assert name not in form_extensions, (
            f"A form extension named {name} already exists"
        )

        form_extensions[name] = cls

    def create(self):
        raise NotImplementedError


class Extendable(object):
    """ Models extending their form classes use this mixin to create the
    extended forms. It also serves as a marker to possibly keep track of all
    classes that use extended forms.

    """

    def extend_form_class(self, form_class, extensions):
        if not extensions:
            return form_class

        for extension in extensions:
            if extension not in form_extensions:
                raise KeyError(f"Unknown form extension: {extension}")

            form_class = form_extensions[extension](form_class).create()

        return form_class



import inspect

from onegov.core.html import sanitize_html
from onegov.core.utils import binary_to_dictionary
from onegov.file.utils import IMAGE_MIME_TYPES_AND_SVG
from onegov.form.widgets import MultiCheckboxWidget
from onegov.form.widgets import OrderedMultiCheckboxWidget
from onegov.form.widgets import TagsWidget
from onegov.form.widgets import UploadWidget
from wtforms import FileField
from wtforms import SelectMultipleField
from wtforms import StringField
from wtforms import TextAreaField
from wtforms import widgets
from wtforms.validators import DataRequired, InputRequired


class MultiCheckboxField(SelectMultipleField):

    widget = MultiCheckboxWidget()
    contains_labels = True

    def __init__(self, *args, **kwargs):
        kwargs['option_widget'] = widgets.CheckboxInput()
        super().__init__(*args, **kwargs)


class OrderedMultiCheckboxField(MultiCheckboxField):
    widget = OrderedMultiCheckboxWidget()


class UploadField(FileField):
    """ A custom file field that turns the uploaded file into a compressed
    base64 string together with the filename, size and mimetype.

    """

    widget = UploadWidget()

    @property
    def data(self):
        caller = inspect.currentframe().f_back.f_locals.get('self')

        # give the required validators the idea that the data is there
        # when the action was to keep the current file - an evil approach
        if isinstance(caller, (DataRequired, InputRequired)):
            truthy = (
                getattr(self, '_data', None) or
                getattr(self, 'action', None) == 'keep'
            )

            return truthy

        return getattr(self, '_data', None)

    @data.setter
    def data(self, value):
        self._data = value

    @property
    def is_image(self):
        return self.data and\
            self.data.get('mimetype') in IMAGE_MIME_TYPES_AND_SVG

    def process_formdata(self, valuelist):
        # the upload widget optionally includes an action with the request,
        # indicating if the existing file should be replaced, kept or deleted
        if valuelist:
            if len(valuelist) == 2:
                self.action, fieldstorage = valuelist
            else:
                self.action = 'replace'
                fieldstorage = valuelist[0]

            if self.action == 'replace':
                self.data = self.process_fieldstorage(fieldstorage)
            elif self.action == 'delete':
                self.data = {}
            elif self.action == 'keep':
                pass
            else:
                raise NotImplementedError()
        else:
            self.data = {}

    def process_fieldstorage(self, fs):
        self.file = getattr(fs, 'file', getattr(fs, 'stream', None))
        self.filename = getattr(fs, 'filename', None)

        if not self.file:
            return {}

        self.file.seek(0)

        try:
            return binary_to_dictionary(self.file.read(), self.filename)
        finally:
            self.file.seek(0)


class HtmlField(TextAreaField):
    """ A textfield with html with integrated sanitation. """

    def __init__(self, *args, **kwargs):
        self.form = kwargs.get('_form')

        if 'render_kw' not in kwargs or not kwargs['render_kw'].get('class_'):
            kwargs['render_kw'] = kwargs.get('render_kw', {})
            kwargs['render_kw']['class_'] = 'editor'

        super().__init__(*args, **kwargs)

    def pre_validate(self, form):
        self.data = sanitize_html(self.data)


class TagsField(StringField):
    """ A tags field for use in conjunction with this library:

    https://github.com/developit/tags-input

    """

    widget = TagsWidget()

    def process_formdata(self, valuelist):
        if valuelist:
            values = (v.strip() for v in valuelist[0].split(','))
            values = (v for v in values if v)

            self.data = list(values)
        else:
            self.data = []

    def process_data(self, value):
        self.data = value and ','.join(value) or ''



import re

from decimal import Decimal
from onegov.form.utils import decimal_range
from pyparsing import (
    alphanums,
    Combine,
    FollowedBy,
    Group,
    Literal,
    MatchFirst,
    nums,
    OneOrMore,
    Optional,
    ParserElement,
    Regex,
    Suppress,
    White,
    Word,
)


# we want newlines to be significant
ParserElement.setDefaultWhitespaceChars(' \t')

unicode_characters = ''.join(
    chr(c) for c in range(65536) if not chr(c).isspace())
text = Word(unicode_characters)
numeric = Word(nums)


def text_without(characters):
    """ Returns all printable text without the given characters. """
    return Word(unicode_characters, excludeChars=characters)


def matches(character):
    """ Returns true if the given character matches the token. """
    return lambda tokens: tokens and tokens[0] == character


def literal(text):
    """" Returns the given value, ignoring the tokens alltogether. """
    return lambda tokens: text


def as_int(tokens):
    """ Converts the token to int if possible. """
    return int(tokens[0]) if tokens else None


def as_decimal(tokens):
    """ Converts the token to decimal if possible. """
    if tokens and tokens[0] == '-':
        tokens = tokens[1:]
        prefix = '-'
    else:
        prefix = '+'

    return Decimal(prefix + '.'.join(tokens)) if tokens else None


def as_uppercase(tokens):
    """ Converts the token to uppercase if possible. """
    return ''.join(tokens).upper() if tokens else None


def as_integer_range(tokens):
    """ Converts the token to an integer range if possible. """
    if tokens:
        return range(int(tokens[0]), int(tokens[1]))


def as_decimal_range(tokens):
    """ Converts the token to a decimal range if possible. """
    if tokens:
        return decimal_range(tokens[0], tokens[1])


def as_regex(tokens):
    """ Converts the token to a working regex if possible. """
    if tokens:
        return re.compile(tokens[0])


def rstrip(tokens):
    """ Strips whitetext on the right of the token. """
    return tokens[0].rstrip()


def unwrap(tokens):
    """ Unwraps grouped tokens. """
    return tokens[0]


def tag(**tags):
    """ Takes the given tags and applies them to the token. """
    def apply_tags(tokens):
        for key, value in tags.items():
            tokens[key] = value
    return apply_tags


def with_whitespace_inside(expr):
    """ Returns an expression that allows for whitespace inside, but not
    outside the expression.

    """
    return Combine(OneOrMore(expr | White(' ', max=1) + expr))


def enclosed_in(expr, characters):
    """ Wraps the given expression in the given characters. """
    left, right = characters
    return Suppress(left) + expr + Suppress(right)


def number_enclosed_in(characters):
    """ Wraps numers in the given characters, making sure the result is an int.

    """
    left, right = characters
    return enclosed_in(numeric, characters).setParseAction(as_int)


def mark_enclosed_in(characters):
    """ Returns a mark (x) inclosed in the given characters. For example,
    ``mark_enclosed_in('[]')`` creates an expression that accepts any of these
    and sets the result of the 'x' value to True, the others to False::

        [x]
        [ ]
    """
    left, right = characters
    return MatchFirst((
        Literal(left + 'x' + right).setParseAction(literal(True)),
        Literal(left + ' ' + right).setParseAction(literal(False))
    ))


def textfield():
    """ Returns a textfield parser.

    Example::

        ____[50]

    The number defines the maximum length.

    Additionally, a regex may be specified to validate the field::

        ___/[0-9]{4}

    """
    length = number_enclosed_in('[]')('length')

    regex = Word(unicode_characters).setParseAction(as_regex)('regex')
    regex = Suppress('/') + regex

    textfield = Suppress('___') + Optional(length) + Optional(regex)
    textfield.setParseAction(tag(type='text'))

    return textfield


def textarea():
    """ Returns a textarea parser.

    Example::

        ...[5]

    The number defines the number of rows.

    """

    rows = number_enclosed_in('[]')('rows')

    textarea = Suppress('...') + Optional(rows)
    textarea.setParseAction(tag(type='textarea'))

    return textarea


def password():
    """ Returns a password field parser.

    Example::

        ***

    """

    return Suppress('***').setParseAction(tag(type='password'))


def email():
    """ Returns an email field parser.

    Example::

        @@@

    """
    return Suppress('@@@').setParseAction(tag(type='email'))


def url():
    """ Returns an url field parser.

    Example::

        http://
        https://

    """
    return Suppress(Regex(r'https?://')).setParseAction(tag(type='url'))


def date():
    """ Returns a date parser.

    Example::

        YYYY.MM.DD

    """

    return Suppress('YYYY.MM.DD').setParseAction(tag(type='date'))


def datetime():
    """ Returns a datetime parser.

    Example::

        YYYY.MM.DD HH:MM

    """

    return Suppress('YYYY.MM.DD HH:MM').setParseAction(tag(type='datetime'))


def time():
    """ Returns a time parser.

    Example::

        HH:MM

    """

    return Suppress('HH:MM').setParseAction(tag(type='time'))


def stdnum():
    """ Returns an stdnum parser.

    Example::

        # iban

    """
    prefix = Suppress('#') + Optional(White(" "))
    parser = prefix + Regex(r'[a-z\.]+')('format')
    parser.setParseAction(tag(type='stdnum'))

    return parser


def fileinput():
    """ Returns a fileinput parser.

    For all kindes of files::
        *.*

    For specific files:
        *.pdf|*.doc
    """
    any_extension = Suppress('*.*')
    some_extension = Suppress('*.') + Word(alphanums) + Optional(Suppress('|'))

    def extract_file_types(tokens):
        tokens['type'] = 'fileinput'
        if len(tokens[0]) == 0:
            tokens['extensions'] = ['*']
        else:
            tokens['extensions'] = [ext.lower() for ext in tokens[0].asList()]

    parser = Group(any_extension | OneOrMore(some_extension))
    parser.setParseAction(extract_file_types)

    return parser


def decimal():
    """ Returns a decimal parser.

    Decimal point is '.'.

    For example:

        0.00
        123
        11.1
        -10.0

    """

    return (Optional('-') + numeric + Optional(Suppress('.') + numeric))\
        .setParseAction(as_decimal)('amount')


def range_field(value_expression, parse_action, type):
    """ Generic range field parser. """

    r = (value_expression + Suppress('..') + value_expression)
    r = r.setParseAction(parse_action)
    r = r.addParseAction(tag(type=type))

    return r


def integer_range_field():
    """ Returns an integer range parser. """

    number = Combine(Optional('-') + numeric)
    return range_field(number, as_integer_range, 'integer_range')


def decimal_range_field():
    """ Returns a decimal range parser. """

    number = Combine(Optional('-') + numeric + Literal('.') + numeric)
    return range_field(number, as_decimal_range, 'decimal_range')


def currency():
    """ Returns a currency parser.

    For example:

        chf
        USD
        Cny

    """

    return Regex(r'[a-zA-Z]{3}').setParseAction(as_uppercase)('currency')


def marker_box(characters):
    """ Returns a marker box:

    Example::

        (x) Male
        [x] Female
        {x} What?
    """

    check = mark_enclosed_in(characters)('checked')
    pricing = enclosed_in(decimal() + currency(), '()')('pricing')

    label_text = with_whitespace_inside(text_without(characters + '()'))
    label = MatchFirst((
        label_text + FollowedBy(pricing),
        Combine(label_text + with_whitespace_inside(text)),
        label_text
    ))('label')

    return check + label + Optional(pricing)


def radio():
    """ Returns a radio parser:

    Example::
        (x) Male
        ( ) Female
    """
    return marker_box('()').setParseAction(tag(type='radio'))


def checkbox():
    """ Returns a checkbox parser:

    Example::
        [x] Male
        [ ] Female
    """
    return marker_box('[]').setParseAction(tag(type='checkbox'))


def fieldset_title():
    """ A fieldset title parser. Fieldset titles are just like headings in
    markdown::

        # My header

    It's possible to have an empty fieldset title (to disable a fieldset):

        # ...

    """

    label = with_whitespace_inside(text).setResultsName('label')

    fieldset_title = Suppress('#') + (Suppress('...') | label)
    fieldset_title = fieldset_title.setParseAction(tag(type='fieldset'))

    return fieldset_title


def field_identifier():
    """ Returns a field identifier parser:

    Example::

        My field *

    """

    required = Literal('*').setParseAction(matches('*'))
    required = Optional(required, default=False)('required')

    # a field name can contain any kind of character, except for a '=' and
    # a '*', which we'll need later for the field definition
    characters = with_whitespace_inside(text_without('*='))
    label = Combine(OneOrMore(characters))('label')

    # a field declaration begins with spaces (so we can differentiate between
    # text and catual fields), then includes the name and the '*' which marks
    # required fields
    return label + required + Suppress('=')



import html

from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import JSON, UUID
from onegov.core.orm.types import UTCDateTime
from onegov.file import AssociatedFiles, File
from onegov.form.display import render_field
from onegov.form.parser import parse_form
from onegov.form.utils import extract_text_from_html, hash_definition
from onegov.form.extensions import Extendable
from onegov.pay import Payable
from onegov.pay import process_payment
from sedate import utcnow
from sqlalchemy import Column, Enum, ForeignKey, Text
from sqlalchemy_utils import observes
from uuid import uuid4
from wtforms import StringField, TextAreaField
from wtforms.fields.html5 import EmailField


class FormSubmission(Base, TimestampMixin, Payable, AssociatedFiles,
                     Extendable):
    """ Defines a submitted form in the database. """

    __tablename__ = 'submissions'

    #: id of the form submission
    id = Column(UUID, primary_key=True, default=uuid4)

    #: name of the form this submission belongs to
    name = Column(Text, ForeignKey("forms.name"), nullable=True)

    #: the title of the submission, generated from the submitted fields
    #: NULL for submissions which are not complete
    title = Column(Text, nullable=True)

    #: the e-mail address associated with the submitee, generated from the
    # submitted fields (may be NULL, even for complete submissions)
    email = Column(Text, nullable=True)

    #: the source code of the form at the moment of submission. This is stored
    #: alongside the submission as the original form may change later. We
    #: want to keep the old form around just in case.
    definition = Column(Text, nullable=False)

    #: the exact time this submissions was changed from 'pending' to 'complete'
    received = Column(UTCDateTime, nullable=True)

    #: the checksum of the definition, forms and submissions with matching
    #: checksums are guaranteed to have the exact same definition
    checksum = Column(Text, nullable=False)

    #: metadata about this submission
    meta = Column(JSON, nullable=False)

    #: the submission data
    data = Column(JSON, nullable=False)

    #: the state of the submission
    state = Column(
        Enum('pending', 'complete', name='submission_state'),
        nullable=False
    )

    #: payment options -> copied from the dfinition at the moment of
    #: submission. This is stored alongside the submission as the original
    #: form setting may change later.
    payment_method = Column(Text, nullable=False, default='manual')

    __mapper_args__ = {
        "polymorphic_on": 'state'
    }

    @property
    def form_class(self):
        """ Parses the form definition and returns a form class. """

        return self.extend_form_class(
            parse_form(self.definition), self.meta.get('extensions'))

    @property
    def form_obj(self):
        """ Returns a form instance containing the submission data. """

        return self.form_class(data=self.data)

    def get_email_field_data(self, form=None):
        form = form or self.form_obj

        email_fields = form.match_fields(
            include_classes=(EmailField, ),
            required=True,
            limit=1
        )
        email_fields += form.match_fields(
            include_classes=(EmailField, ),
            required=False,
            limit=1
        )

        if email_fields:
            return form._fields[email_fields[0]].data

    @observes('definition')
    def definition_observer(self, definition):
        self.checksum = hash_definition(definition)

    @observes('state')
    def state_observer(self, state):
        if self.state == 'complete':
            form = self.form_class(data=self.data)

            title_fields = form.match_fields(
                include_classes=(StringField, ),
                exclude_classes=(TextAreaField, ),
                required=True,
                limit=3
            )

            if title_fields:
                self.title = extract_text_from_html(', '.join(
                    html.unescape(render_field(form._fields[id]))
                    for id in title_fields
                ))

            if not self.email:
                self.email = self.get_email_field_data(form=form)

            # only set the date the first time around
            if not self.received:
                self.received = utcnow()

    def process_payment(self, price, provider=None, token=None):
        """ Takes a request, optionally with the provider and the token
        by the provider that can be used to charge the credit card and creates
        a payment record if necessary.

        Returns True or a payment object if the payment was processed
        successfully. That is, if there is a payment or if there is no payment
        required the method returns truthy.

        """

        if price and price.amount > 0:
            return process_payment(self.payment_method, price, provider, token)

        return True


class PendingFormSubmission(FormSubmission):
    __mapper_args__ = {'polymorphic_identity': 'pending'}


class CompleteFormSubmission(FormSubmission):
    __mapper_args__ = {'polymorphic_identity': 'complete'}


class FormFile(File):
    __mapper_args__ = {'polymorphic_identity': 'formfile'}



import pytest

from onegov.form import FormCollection, FormExtension
from webob.multidict import MultiDict
from wtforms import ValidationError


def test_has_submissions(session):
    collection = FormCollection(session)
    form = collection.definitions.add('Newsletter', definition="E-Mail = @@@")

    assert not form.has_submissions()

    data = MultiDict([
        ('e_mail', 'test@example.org'),
    ])

    collection.submissions.add(
        'newsletter', form.form_class(data), state='complete')

    assert not form.has_submissions(with_state='pending')
    assert form.has_submissions(with_state='complete')
    assert form.has_submissions()


def test_form_extensions(session):
    collection = FormCollection(session)

    members = collection.definitions.add('Members', definition="E-Mail = @@@")

    class CorporateOnly(object):

        def validate_e_mail(self, value):
            # note, you probably don't want to do this in a real world
            # project as the name of the e-mail field might change and
            # the validation would not be triggered!
            if not value.data.endswith('seantis.ch'):
                raise ValidationError("Only seantis.ch e-mails are allowed")

    class CorporateOnlyExtension(FormExtension, name='corporate-emails-only'):

        def create(self):
            class ExtendedForm(self.form_class, CorporateOnly):
                pass

            return ExtendedForm

    members.meta['extensions'] = ['corporate-emails-only']
    assert issubclass(members.form_class, CorporateOnly)

    session.flush()

    # the validation should fail here
    data = MultiDict([
        ('e_mail', 'test@example.org'),
    ])

    form = members.form_class(data)
    form.validate()
    assert form.errors

    with pytest.raises(AssertionError) as e:
        collection.submissions.add('newsletter', form, state='complete')

    assert "the given form doesn't validate" in str(e)

    # now it should work
    data = MultiDict([
        ('e_mail', 'test@seantis.ch'),
    ])

    form = members.form_class(data)
    form.validate()
    assert not form.errors

    # the extensions are carried over to the submission
    submission = collection.submissions.add(
        'members', form, state='complete')
    assert issubclass(submission.form_class, CorporateOnly)



import humanize

from cgi import escape
from onegov.form import _
from wtforms.widgets import ListWidget, FileInput, TextInput
from wtforms.widgets.core import HTMLString


class OrderedListWidget(ListWidget):
    """ Extends the default list widget with automated ordering using the
    translated text of each element.

    """

    def __call__(self, field, **kwargs):

        # ListWidget expects a field internally, but it will only use
        # its id property and __iter__ method, so we can get away
        # with passing a fake field with an id and an iterator.
        #
        # It's not great, since we have to assume internal knowledge,
        # but builting a new field or changing the existing one would
        # require even more knowledge, so this is the better approach
        #
        # We also need to call each field once so it gets hooked up with
        # our translation machinary
        ordered = [subfield for subfield in field]
        ordered.sort(key=lambda f: (f(), str(f.label.text))[1])

        class FakeField(object):

            id = field.id

            def __iter__(self):
                return iter(ordered)

        return super().__call__(FakeField(), **kwargs)


class MultiCheckboxWidget(ListWidget):
    """ The default list widget with the label behind the checkbox. """

    def __init__(self, *args, **kwargs):
        kwargs['prefix_label'] = False
        super().__init__(*args, **kwargs)


class OrderedMultiCheckboxWidget(MultiCheckboxWidget, OrderedListWidget):
    """ The sorted list widget with the label behind the checkbox. """
    pass


class CoordinateWidget(TextInput):
    """ Widget containing the coordinates for the
    :class:`onegov.form.fields.CoordinateField` class.

    Basically a text input with a class. Meant to be enhanced on the browser
    using javascript.

    """

    def __call__(self, field, **kwargs):
        kwargs['class_'] = (kwargs.get('class_', '') + ' coordinate').strip()
        return super().__call__(field, **kwargs)


class UploadWidget(FileInput):
    """ An upload widget for the :class:`onegov.form.fields.UploadField` class,
    which supports keeping, removing and replacing already uploaded files.

    This is necessary as file inputs are read-only on the client and it's
    therefore rather easy for users to lose their input otherwise (e.g. a
    form with a file is rejected because of some mistake - the file disappears
    once the response is rendered on the client).

    """

    def __call__(self, field, **kwargs):
        force_simple = kwargs.pop('force_simple', False)
        input_html = super().__call__(field, **kwargs)

        if force_simple or field.errors or not field.data:
            return HTMLString("""
                <div class="upload-widget without-data">
                    {}
                </div>
            """.format(input_html))
        else:
            return HTMLString("""
                <div class="upload-widget with-data">
                    <p>{existing_file_label}: {filename} ({filesize}) ✓</p>
                    <ul>
                        <li>
                            <input type="radio" id="{name}-0" name="{name}"
                                   value="keep" checked="">
                            <label for="{name}-0">{keep_label}</label>
                        </li>
                        <li>
                            <input type="radio" id="{name}-1" name="{name}"
                                   value="delete">
                            <label for="{name}-1">{delete_label}</label>
                        </li>
                        <li>
                            <input type="radio" id="{name}-2" name="{name}"
                                   value="replace">
                            <label for="{name}-2">{replace_label}</label>
                            <div>
                                <label>
                                    <div data-depends-on="{name}/replace"
                                         data-hide-label="false">
                                        {input_html}
                                    </div>
                                </label>
                            </div>
                        </li>
                    </ul>
                </div>
            """.format(
                # be careful, we do our own html generation here without any
                # safety harness - we need to carefully escape values the user
                # might supply
                filesize=humanize.naturalsize(field.data['size']),
                filename=escape(field.data['filename'], quote=True),
                name=field.id,
                input_html=input_html,
                existing_file_label=field.gettext(_('Uploaded file')),
                keep_label=field.gettext(_('Keep file')),
                delete_label=field.gettext(_('Delete file')),
                replace_label=field.gettext(_('Replace file'))
            ))


class TagsWidget(TextInput):
    # for use with https://github.com/developit/tags-input
    input_type = 'tags'



from datetime import datetime, time
from onegov.form import render_field


class MockField(object):

    def __init__(self, type, data):
        self.type = type
        self.data = data

        if isinstance(data, str):
            self.choices = [(data, data)]
        elif isinstance(data, list):
            self.choices = [(c, c) for c in data]


def test_render_textfields():
    assert render_field(MockField('StringField', 'asdf')) == 'asdf'
    assert render_field(MockField('StringField', '<b>')) == '&lt;b&gt;'


def test_render_password():
    assert render_field(MockField('PasswordField', '123')) == '***'
    assert render_field(MockField('PasswordField', '1234')) == '****'
    assert render_field(MockField('PasswordField', '12345')) == '*****'


def test_render_date_field():
    assert render_field(MockField('DateField', datetime(1984, 4, 6)))\
        == '06.04.1984'
    assert render_field(MockField('DateTimeLocalField', datetime(1984, 4, 6)))\
        == '06.04.1984 00:00'
    assert render_field(MockField('TimeField', time(10, 0))) == '10:00'


def test_render_upload_field():
    assert render_field(MockField('UploadField', {
        'filename': '<b.txt>', 'size': 1000
    })) == '&lt;b.txt&gt; (1.0 kB)'


def test_render_radio_field():
    assert render_field(MockField('RadioField', 'selected')) == '✓ selected'


def test_render_multi_checkbox_field():
    assert render_field(MockField('MultiCheckboxField', ['a', 'b']))\
        == '✓ a<br>✓ b'



import re

from decimal import Decimal
from hashlib import md5
from unidecode import unidecode

_unwanted_characters = re.compile(r'[^a-zA-Z0-9]+')
_html_tags = re.compile(r'<.*?>')


def as_internal_id(label):
    clean = unidecode(label).strip(' ').lower()
    clean = _unwanted_characters.sub('_', clean)

    return clean


def get_fields_from_class(cls):
    fields = []

    for name in dir(cls):
        if not name.startswith('_'):
            field = getattr(cls, name)

            if hasattr(field, '_formfield'):
                fields.append((name, field))

    fields.sort(key=lambda x: (x[1].creation_counter, x[0]))

    return fields


def extract_text_from_html(html):
    return _html_tags.sub('', html)


class decimal_range(object):
    """ Implementation of Python's range builtin using decimal values instead
    of integers.

    """

    def __init__(self, start, stop, step=None):
        if step is None and start <= stop:
            step = '1.0'
        elif step is None and start >= stop:
            step = '-1.0'

        self.start = self.current = Decimal(start)
        self.stop = Decimal(stop)
        self.step = Decimal(step)

        assert self.step != Decimal('0')

    def __repr__(self):
        if self.start <= self.stop and self.step == Decimal('1.0'):
            return "decimal_range('{}', '{}')".format(self.start, self.stop)
        elif self.start >= self.stop and self.step == Decimal('-1.0'):
            return "decimal_range('{}', '{}')".format(self.start, self.stop)
        else:
            return "decimal_range('{}', '{}', '{}')".format(
                self.start, self.stop, self.step
            )

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        return (self.start, self.stop, self.step)\
            == (other.start, other.stop, other.step)

    def __iter__(self):
        return self

    def __next__(self):
        result, self.current = self.current, self.current + self.step

        if self.step > 0 and result >= self.stop:
            raise StopIteration

        if self.step < 0 and result <= self.stop:
            raise StopIteration

        return result


def hash_definition(definition):
    return md5(definition.encode('utf-8')).hexdigest()
